@@ -1,11 +1,16 @@
 import argparse
+import os
 from pathlib import Path
+import json
 import wsiprocess as wp
 
 
 class Args:
     def __init__(self, command):
         self.build_args(command)
+        self.fillattrs(keys=[
+            "annotation", "rule", "export_thumbs", "on_annotation", "minmax",
+            "crop_bbox", "extract_foreground"])
 
     def set_base_parser(self):
         self.base_parser = argparse.ArgumentParser(
@@ -40,6 +45,15 @@ class Args:
         parser.add_argument(
             "-dh", "--dot_bbox_height", type=int,
             help="Height of bbox translated from dot annotation.")
+        parser.add_argument(
+            "-ex", "--ext", type=str, default="png",
+            help="Extension of extracted patches")
+        parser.add_argument(
+            "-ma", "--magnification", type=int,
+            help="Magnification of extracted patches")
+        parser.add_argument(
+            "-mw", "--max_workers", type=int, default=os.cpu_count()//2,
+            help="The maximum number of workers to use in patching.")
         parser.add_argument(
             "-ss", "--start_sample", action="store_true",
             help="Generate samples at the start of the process.")
@@ -55,6 +69,9 @@ class Args:
         parser.add_argument(
             "-ve", "--verbose", action="store_true",
             help="Show progress bar while patching.")
+        parser.add_argument(
+            "-dr", "--dryrun", action="store_true",
+            help="Run patching only for first 100 patches.")
 
     def set_wsi_arg(self, parser):
         parser.add_argument(
@@ -65,10 +82,6 @@ class Args:
         parser.add_argument(
             "annotation", type=str, default=False,
             help="Path to the annotation file.")
-        parser.add_argument(
-            "-ma", "--magnification", choices={40, 20, 10}, default=40,
-            type=int,
-            help="Magnification to process.")
         parser.add_argument(
             "-ru", "--rule", type=Path,
             help="File to define the inclusion / exclusion relationship.")
@@ -88,7 +101,7 @@ class Args:
     def add_on_annotation(self, parser, slide_is_sparse=False):
         on_annotation_param = 0.0001 if slide_is_sparse else 0.01
         parser.add_argument(
-            "-oa", "--on_annotation", type=float, default=on_annotation_param,
+            "-oa", "--on_annotation", type=json.loads, default=on_annotation_param,
             help="The ratio of overlapped area of a patch and the annotated.")
 
     def add_binarization_method(self, parser):
@@ -171,32 +184,24 @@ class Args:
 
         self.base_parser.parse_args(command, namespace=self)
 
+    def fillattr(self, key: str):
+        if not hasattr(self, key):
+            setattr(self, key, False)
+
+    def fillattrs(self, keys):
+        for key in keys:
+            self.fillattr(key)
+
 
 def process_annotation(args, slide, rule):
-    if args.method == "evaluation":
-        annotation = wp.annotation("")
-        annotation.dot_to_bbox(args.dot_bbox_width, args.dot_bbox_height)
-        if args.minmax:
-            min_, max_ = map(int, args.minmax.split("-"))
-            annotation.make_masks(
-                slide, foreground="minmax", min_=min_, max_=max_)
-        else:
-            annotation.make_masks(
-                slide, foreground="otsu")
-
+    annotation = wp.annotation(args.annotation, slidename=slide.filename)
+    annotation.dot_to_bbox(args.dot_bbox_width, args.dot_bbox_height)
+    if args.minmax:
+        min_, max_ = map(int, args.minmax.split("-"))
+        annotation.make_masks(
+            slide, rule, foreground_fn="minmax", min_=min_, max_=max_)
     else:
-        annotation = wp.annotation(args.annotation, slidename=slide.filename)
-        annotation.dot_to_bbox(args.dot_bbox_width, args.dot_bbox_height)
-        if hasattr(args, "minmax") and args.minmax:
-            min_, max_ = map(int, args.minmax.split("-"))
-            annotation.make_masks(
-                slide, rule, foreground="minmax", min_=min_, max_=max_)
-        else:
-            annotation.make_masks(slide, rule, foreground=True)
-
-    if hasattr(args, "extract_foreground"):
-        if not (args.extract_foreground and "foreground" in annotation.classes):
-            annotation.classes.remove("foreground")
+        annotation.make_masks(slide, rule, foreground_fn="otsu")
 
     return annotation
 
@@ -204,20 +209,17 @@ def process_annotation(args, slide, rule):
 def main(command=None):
     args = Args(command)
     slide = wp.slide(args.wsi)
-    rule = wp.rule(args.rule) if hasattr(args, "rule") and args.rule else False
+    rule = wp.rule(args.rule) if args.rule else False
     annotation = process_annotation(args, slide, rule)
 
-    if hasattr(args, "export_thumbs") and args.export_thumbs:
+    if args.export_thumbs:
         thumbs_dir = args.save_to/slide.filestem/"thumbs"
-        if not thumbs_dir.exists():
-            thumbs_dir.mkdir(parents=True)
+        thumbs_dir.mkdir(parents=True, exist_ok=True)
         annotation.export_thumb_masks(thumbs_dir)
 
-    if args.method == "evaluation":
-        on_annotation = False
-    else:
-        on_annotation = args.on_annotation
-    crop_bbox = args.crop_bbox if hasattr(args, "crop_bbox") else False
+    extract_classes = annotation.classes
+    if not (args.extract_foreground and "foreground" in annotation.classes):
+        extract_classes.remove("foreground")
 
     patcher = wp.patcher(
         slide, args.method,
@@ -228,15 +230,20 @@ def main(command=None):
         overlap_width=args.overlap_width,
         overlap_height=args.overlap_height,
         on_foreground=args.on_foreground,
-        on_annotation=on_annotation,
+        on_annotation=args.on_annotation,
         offset_x=args.offset_x,
         offset_y=args.offset_y,
+        ext=args.ext,
+        magnification=args.magnification,
         start_sample=args.start_sample,
         finished_sample=args.finished_sample,
         no_patches=args.no_patches,
-        crop_bbox=crop_bbox,
-        verbose=args.verbose)
-    patcher.get_patch_parallel(annotation.classes)
+        crop_bbox=args.crop_bbox,
+        verbose=args.verbose,
+        dryrun=args.dryrun)
+
+    patcher.get_patch_parallel(
+        extract_classes, max_workers=args.max_workers)
 
     if args.method == "detection":
         converter = wp.converter(
