@@ -2,15 +2,18 @@
 """Patcher object to extract patches from whole slide images.
 """
 
+import warnings
 import random
 from itertools import product
 import json
+import os
+from concurrent.futures import ThreadPoolExecutor, as_completed
 
-from joblib import Parallel, delayed
 from tqdm import tqdm
 import numpy as np
 import cv2
 from pathlib import Path
+import pandas as pd
 
 from .verify import Verify
 
@@ -36,8 +39,10 @@ class Patcher:
         offset_y (int, optional): The offset pixels along the y-axis.
         on_foreground (float, optional): Ratio of overlap area between patches
             and foreground area.
-        on_annotation (float, optional): Ratio of overlap area between patches
-            and annotation.
+        on_annotation (float or dict): Ratio of overlap area between patches
+            and annotation. dict is also available ex: {"label":value}.
+        magnification (int, optional): Magnification of output patches.
+        ext (str, optional): Extension of extracted patches.
         start_sample (bool, optional): Whether to save sample patches on
             Patcher starting.
         finished_sample (bool, optional): Whether to save sample patches on
@@ -47,6 +52,7 @@ class Patcher:
         no_patches (bool, optional): If set, Patcher runs without extracting
             patches and saves them to disk.
         verbose (bool, optional): If set, a progress bar appears when patching.
+        dryrun (bool, optional): Only run patching for first 100 patches.
 
     Attributes:
         slide (wsiprocess.slide.Slide): Slide object.
@@ -61,8 +67,10 @@ class Patcher:
         masks (dict): Masks to show the location of classes.
         classes (list): Classes to extract.
         save_to (str): The root of the output directory.
-        p_width (int): The width of the output patches.
-        p_height (int): The height of the output patches.
+        p_width (int): The width of the cropped patches. It magnification is
+            not set, it is same as output patches.
+        p_height (int): The height of the cropped patches. It magnification is
+            not set, it is same as output patches.
         p_area (int): The area of single patch.
         o_width (int): The width of the overlap areas of patches.
         o_height (int): The height of the overlap areas of patches.
@@ -70,14 +78,18 @@ class Patcher:
         offset_y (int): The The offset pixels along the y-axis.
         on_foreground (float): Ratio of overlap area between patches and
             foreground area.
-        on_annotation (float): Ratio of overlap area between patches and
-            annotation.
+        on_annotation (float or dict): Ratio of overlap area between patches
+            and annotation. dict is also available ex: {"label":value}.
+        magnification (int): Magnification of output patches.
+        p_scale (float): Ratio of p_width or p_height to output patch size.
+        ext (str): Extension of extracted patches.
         start_sample (bool): Whether to save sample patches on Patcher start.
         finished_sample (bool): Whether to save sample patches on Patcher
             finish.
         extract_patches (bool): Whether to save patches when Patcher runs.
         no_patches (bool): Whether to save patches when Patcher runs.
         verbose (bool, optional): If set, a progress bar appears when patching.
+        dryrun (bool, optional): Only run patching for first 100 patches.
 
         x_lefttop (list): Offsets of patches to the x-axis direction except for
             the right edge.
@@ -94,8 +106,9 @@ class Patcher:
             self, slide, method, annotation=False, save_to=".",
             patch_width=256, patch_height=256, overlap_width=0,
             overlap_height=0, offset_x=0, offset_y=0, on_foreground=0.5,
-            on_annotation=0.5, start_sample=False, finished_sample=False,
-            no_patches=False, crop_bbox=False, verbose=False):
+            on_annotation=0.5, ext="jpg", magnification=False,
+            start_sample=False, finished_sample=False, no_patches=False,
+            crop_bbox=False, verbose=False, dryrun=False):
         self.verify = Verify(
             save_to, slide.filestem, method, start_sample, finished_sample,
             no_patches, crop_bbox)
@@ -112,9 +125,10 @@ class Patcher:
         self.method = method.lower()
         self.wsi_width = slide.width
         self.wsi_height = slide.height
-        self.p_width = int(patch_width)
-        self.p_height = int(patch_height)
-        self.p_area = patch_width * patch_height
+        self.set_magnification(slide, magnification)
+        self.p_width = int(patch_width*self.p_scale)
+        self.p_height = int(patch_height*self.p_scale)
+        self.p_area = self.p_width * self.p_height
         self.o_width = int(overlap_width)
         self.o_height = int(overlap_height)
         self.offset_x = int(offset_x)
@@ -133,7 +147,10 @@ class Patcher:
         self.last_x = self.slide.width - patch_width
         self.last_y = self.slide.height - patch_height
 
-        self.get_iterator()
+        self.dryrun = dryrun
+        self.get_iterator(dryrun)
+
+        self.ext = ext
 
         self.start_sample = start_sample
         self.finished_sample = finished_sample
@@ -146,10 +163,14 @@ class Patcher:
         if annotation:
             self.masks = annotation.masks
             self.classes = annotation.classes
-            self.on_annotation = on_annotation
+            if isinstance(on_annotation, (float, int)):
+                # if float, apply same threshold value for all classes
+                self.on_annotation = {cl: on_annotation for cl in self.classes}
+            else:
+                self.on_annotation = on_annotation
         else:
             self.masks = False
-            self.classes = ["none"]
+            self.classes = []
             self.on_annotation = False
 
         self.save_to = save_to
@@ -159,7 +180,7 @@ class Patcher:
     def __str__(self):
         return "wsiprocess.patcher.Patcher {}".format(self.slide.path)
 
-    def get_iterator(self):
+    def get_iterator(self, dryrun=False):
         self.iterator = list(product(self.x_lefttop, self.y_lefttop))
         bottomedge = [(x, self.last_y) for x in self.x_lefttop]
         rightedge = [(self.last_x, y) for y in self.y_lefttop]
@@ -168,6 +189,24 @@ class Patcher:
         self.iterator += rightedge
         self.iterator += rightbottom
 
+        if dryrun:
+            self.iterator = self.iterator[:100]
+
+    def set_magnification(self, slide, magnification):
+        self.magnification = magnification
+        if self.magnification:
+            if slide.magnification is None:
+                raise KeyError(
+                    "{} has no magnification property".format(slide.path))
+            if slide.magnification % self.magnification != 0:
+                msg = "magnification={} is not a factor of \
+                    slide.magnification={}"
+                warnings.warn(
+                    msg.format(self.magnification, slide.magnification))
+            self.p_scale = slide.magnification / self.magnification
+        else:
+            self.p_scale = 1
+
     def save_patch_result(self, x, y, cls):
         """Save the extracted patch data to result
 
@@ -443,14 +482,16 @@ class Patcher:
             return []
         else:
             # Find mask coords
-            patch_mask = self.masks[cls][y:y+self.p_height, x:x+self.p_width]
-            mask_png_path = "{}/{}/masks/{}/{:06}_{:06}.png".format(
-                self.save_to, self.filestem, cls, x, y)
-            cv2.imwrite(mask_png_path, patch_mask, (cv2.IMWRITE_PXM_BINARY, 1))
+            patch_mask = self.annotation.get_patch_mask(
+                cls, x, y, self.p_width, self.p_height)
+            mask_path = "{}/{}/masks/{}/{:06}_{:06}.{}".format(
+                self.save_to, self.filestem, cls, x, y, self.ext)
+            if not self.no_patches:
+                cv2.imwrite(mask_path, patch_mask, (cv2.IMWRITE_PXM_BINARY, 1))
             # contours, _ = cv2.findContours(
             #   patch_mask, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_NONE)
             masks = []
-            mask = {"coords": mask_png_path, "class": cls}
+            mask = {"coords": mask_path, "class": cls}
             masks.append(mask)
             return masks
 
@@ -478,6 +519,7 @@ class Patcher:
         self.result["overlap_hegiht"] = self.o_height
         self.result["offset_x"] = self.offset_x
         self.result["offset_y"] = self.offset_y
+        self.result["ext"] = self.ext
         self.result["start_sample"] = self.start_sample
         self.result["finished_sample"] = self.finished_sample
         self.result["no_patches"] = self.no_patches
@@ -485,15 +527,66 @@ class Patcher:
         self.result["on_annotation"] = self.on_annotation
         self.result["dot_bbox_width"] = self.dot_bbox_width
         self.result["dot_bbox_height"] = self.dot_bbox_height
+        self.result["magnification"] = self.magnification
+        self.result["dryrun"] = self.dryrun
         self.result["save_to"] = str(Path(self.save_to).absolute())
         self.result["classes"] = sorted(self.classes)
 
         self.remove_dup_in_results()
 
-        save_as = "{}/{}/results.json".format(self.save_to, self.filestem)
-        with open(save_as, "w") as f:
+        with open(
+            "{}/{}/results.json".format(self.save_to, self.filestem),
+                "w") as f:
             json.dump(self.result, f, indent=4)
 
+        coords = pd.DataFrame(self.result["result"])
+        if not self.result["result"]:
+            return
+
+        coords.sort_values(by=["x", "y"], inplace=True)
+        coords.reset_index(drop=True, inplace=True)
+        if self.method == "classification":
+            for cls in self.classes:
+                coords[cls] = (coords["class"] == cls)
+
+            # aggregate if a patch has multiple classes
+            if coords.duplicated(subset=["x", "y", "w", "h"]).sum() > 0:
+                dup_to_keep = ~coords.duplicated(
+                    subset=["x", "y", "w", "h"], keep="first")
+
+                for i, row in coords[dup_to_keep].iterrows():
+                    dup_x = coords.x == row.x
+                    dup_y = coords.y == row.y
+                    dup_w = coords.w == row.w
+                    dup_h = coords.h == row.h
+                    dup = coords[dup_x & dup_y & dup_w & dup_h]
+                    coords.loc[i, self.classes] = dup[self.classes].any()
+
+                coords.drop_duplicates(
+                    subset=["x", "y", "w", "h"], keep="first", inplace=True)
+
+            coords.drop(columns="class", inplace=True)
+
+        elif self.method == "detection":
+            # column name: bbs
+            # not implemented
+            pass
+
+        elif self.method == "segmentation":
+            # column name: masks
+            for cls in self.classes:
+                coords[cls] = False
+
+            def mask_cls_to_column(x):
+                for mask in x.masks:
+                    coords.loc[x.name, mask["class"]] = True
+
+            coords.apply(mask_cls_to_column, axis=1)
+
+        coords.to_csv(
+            "{}/{}/coords.csv".format(self.save_to, self.filestem),
+            index=None)
+
     def get_patch(self, x, y, classes=False):
         """Extract a single patch.
 
@@ -517,51 +610,58 @@ class Patcher:
             on_annotation_classes = ["foreground"]
         for cls in on_annotation_classes:
             if not self.no_patches:
-                patch = self.slide.crop(
-                    x, y, self.p_width, self.p_height)
-                patch.save(
-                    "{}/{}/patches/{}/{:06}_{:06}.png".format(
-                        self.save_to, self.filestem, cls, x, y))
+                patch = self.crop_patch(x, y)
+                self.save_patch(patch, "{}/{}/patches/{}/{:06}_{:06}.{}".format(
+                    self.save_to, self.filestem, cls, x, y, self.ext))
             self.save_patch_result(x, y, cls)
 
-    def get_patch_parallel(self, classes=False, cores=-1):
+    def get_patch_parallel(self, classes=False, max_workers=-1):
         """Run get_patch() in parallel.
 
         Args:
             classes (list): Classes to extract.
-            cores (int): Threads to run. -1 means same as the number of cores.
+            max_workers (int): Workers to run. -1 runs with cores*5 threads.
         """
+        for cls in classes:
+            assert cls in self.on_annotation, f"on_annotation of {cls} not set"
+
+        if max_workers == 0 | max_workers < -1:
+            msg = "max_workers must be 1 or larger, or -1"
+            msg += f", got {max_workers}"
+            warnings.warn(msg)
         for cls in classes:
             if not self.no_patches:
                 self.verify.make_dir(
                     "{}/{}/patches/{}".format(
                         self.save_to, self.filestem, cls))
-            if self.method == "segmentation":
-                self.verify.make_dir(
-                    "{}/{}/masks/{}".format(self.save_to, self.filestem, cls))
+                if self.method == "segmentation":
+                    self.verify.make_dir(
+                        "{}/{}/masks/{}".format(
+                            self.save_to, self.filestem, cls))
 
         if self.start_sample:
             self.get_random_sample("start", 3)
 
-        def Runner(**joblib_args):
-            def run(**tqdm_args):
-                def tmp(op_iter):
-                    def progress(args): return lambda x: tqdm(x, **args)
-                    bar = progress(tqdm_args)
-                    return Parallel(**joblib_args)(bar(op_iter))
-                return tmp
-            return run
-
-        if self.verbose:
-            run = Runner(n_jobs=cores, backend="threading", verbose=0)
-            runner = run(
-                desc=f"[{self.slide.filename} {self.p_width}x{self.p_height}]")
+        if max_workers == -1:
+            max_workers = os.cpu_count()*5
         else:
-            runner = Parallel(n_jobs=cores, backend="threading", verbose=0)
+            max_workers = max_workers
 
-        # from the left top to just before the right bottom.
-        runner([delayed(self.get_patch)(x, y, classes)
-                for x, y in self.iterator])
+        with ThreadPoolExecutor(max_workers=max_workers) as executor:
+            futures = []
+            for x, y in self.iterator:
+                futures.append(
+                    executor.submit(
+                        self.get_patch,
+                        *(x, y, classes)
+                    )
+                )
+            if self.verbose:
+                desc = f"[{self.filepath} {self.p_width}x{self.p_height}]"
+                [_ for _ in tqdm(
+                    as_completed(futures),
+                    desc=desc,
+                    total=len(futures))]
 
         # save results
         self.save_results()
@@ -577,16 +677,23 @@ class Patcher:
 
         for patch in self.result["result"]:
             for bb in patch["bbs"]:
-                if bb["class"] in classes:
-                    mini_patch = self.slide.crop(
+                if bb["class"] not in classes:
+                    continue
+                # not properly works with the magnification settings
+                mini_patch = self.crop_patch(
+                    patch["x"] + bb["x"],
+                    patch["y"] + bb["y"],
+                    bb["w"],
+                    bb["h"]
+                )
+                self.save_patch(
+                    mini_patch,
+                    "{}/{}/mini_patches/{}/{:06}_{:06}.{}".format(
+                        self.save_to, self.filestem,
+                        bb["class"],
                         patch["x"] + bb["x"],
                         patch["y"] + bb["y"],
-                        bb["w"],
-                        bb["h"])
-                    mini_patch.save(
-                        "{}/{}/mini_patches/{}/{:06}_{:06}.png".format(
-                            self.save_to, self.filestem, bb["class"],
-                            patch["x"] + bb["x"], patch["y"] + bb["y"]))
+                        self.ext))
 
     def patch_on_foreground(self, x, y):
         """Check if the patch is on the foreground area.
@@ -598,8 +705,8 @@ class Patcher:
         Returns:
             (bool): Whether the patch is on the foreground area.
         """
-        patch_mask = self.masks["foreground"][y:y +
-                                              self.p_height, x:x+self.p_width]
+        patch_mask = self.annotation.get_patch_mask(
+            "foreground", x, y, self.p_width, self.p_height)
         return (patch_mask.sum() / self.p_area) >= self.on_foreground
 
     def patch_on_annotation(self, cls, x, y):
@@ -614,8 +721,9 @@ class Patcher:
         Returns:
             (bool): Whether the patch is on the anntation.
         """
-        patch_mask = self.masks[cls][y:y+self.p_height, x:x+self.p_width]
-        return (patch_mask.sum() / self.p_area) >= self.on_annotation
+        patch_mask = self.annotation.get_patch_mask(
+            cls, x, y, self.p_width, self.p_height)
+        return (patch_mask.sum() / self.p_area) >= self.on_annotation[cls]
 
     def get_random_sample(self, phase, sample_count=1):
         """Get random patch to check if the patcher can work properly.
@@ -627,7 +735,27 @@ class Patcher:
         for i in range(sample_count):
             x = random.choice(self.x_lefttop)
             y = random.choice(self.y_lefttop)
-            patch = self.slide.crop(x, y, self.p_width, self.p_height)
-            patch.save(
-                "{}/{}/{}_sample/{:06}_{:06}.png".format(
-                    self.save_to, self.filestem, phase, x, y))
+            patch = self.crop_patch(x, y)
+            self.save_patch(patch, "{}/{}/{}_sample/{:06}_{:06}.{}".format(
+                self.save_to, self.filestem, phase, x, y, self.ext))
+
+    def crop_patch(self, x, y, w=False, h=False):
+        if not w:
+            w = self.p_width
+        if not h:
+            h = self.p_height
+        return self.slide.crop(x, y, w, h)
+
+    def save_patch(self, patch, save_as):
+        if patch.mode == "RGBA" and self.ext == "jpg":
+            warnings.warn(
+                "patch has RGBA data. Discarding alpha to save as jpg")
+            patch = patch.convert("RGB")
+
+        if self.magnification:
+            patch = patch.resize((
+                int(self.p_width//self.p_scale),
+                int(self.p_height//self.p_scale)
+            ))
+
+        patch.save(save_as)
