@@ -4,6 +4,7 @@ Verify class works for verification of the output directory, annotation files,
 rule files, etc. Mainly runs for cli.
 """
 
+import warnings
 from pathlib import Path
 from .error import SizeError, OnParamError
 
@@ -51,6 +52,10 @@ class Verify:
         """Ensure the output directories exists for each tasks.
         """
         base_dir = Path(self.save_to)/self.filestem
+        if base_dir.exists():
+            warnings.warn(
+                "saving results to {}, but it already exists.".format(base_dir)
+            )
         self.make_dir(base_dir)
         if self.method == "evaluation" and not self.no_patches:
             self.make_dir(base_dir/"patches"/"foreground")
@@ -114,11 +119,18 @@ class Verify:
         Raises:
             wsiprocess.error.SizeError: If the sizes are invalid.
         """
-        if not isinstance(on_annotation, bool):
-            if on_annotation <= 0 or on_annotation > 1:
-                raise OnParamError(
-                    "on_annotation is between 0 and 1 excluding 0.")
+        if on_annotation is not False:
+            if isinstance(on_annotation, (float, int)):
+                if on_annotation <= 0 or on_annotation > 1:
+                    raise OnParamError(
+                        "on_annotation is between 0 and 1 excluding 0.")
+            elif isinstance(on_annotation, dict):
+                for v in on_annotation.values():
+                    if v <= 0 or v > 1:
+                        raise OnParamError(
+                            "on_annotation is between 0 and 1 excluding 0.")
 
-        if on_foreground <= 0 or on_foreground > 1:
-            raise OnParamError(
-                "on_foreground is between 0 and 1 excluding 0.")
+        if on_foreground is not False:
+            if on_foreground <= 0 or on_foreground > 1:
+                raise OnParamError(
+                    "on_foreground is between 0 and 1 excluding 0.")
